/*
  
  morph.c - WordNet search code morphology functions
  
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "wn.h"
#include "wnconsts.h"

static char *Id = "$Id: morph.c,v 1.10 92/04/07 14:03:26 wn Exp $";

#define WORDSIZE	256

char  *sufx[] ={ 
/* Noun suffixes */
"s","ses","xes","zes","ches","shes",
/* Verb suffixes */
"s","ies","es","es","ed","ed","ing","ing",
/* Adjective suffixes */
"er","est","er", "est", "ally","ally","ily","ly","ly","ly"
};

char  *addr[] ={ 
/* Noun endings */
"", "s", "x", "z", "ch", "sh", 
/* Verb endings */
"", "y", "e", "", "e", "", "e", "",
/* Adjective endings */
"", "", "e", "e", "al", "", "y", "le", "e", ""
};

int offsets[4] = { 0, 0, 6, 14 };
int cnts[4] = { 0, 6, 8, 10 };
char *exc_fnames[4] = { NULL, "noun.exc", "verb.exc", "adj.exc" };

static int strend(str1,str2)
char *str1,*str2;
{
    char *pt1;

    if(strlen(str2) >= strlen(str1))
	return(0);
    else {
	pt1=str1;
	pt1=strchr(str1,0);
	pt1=pt1-strlen(str2);
	return(!strcmp(pt1,str2));
    }
}

static char *wordbase(word,ender)
char *word;
int ender;
{
    char *pt1;
    static char copy[WORDSIZE];

    strcpy(copy, word);
    if(strend(copy,sufx[ender])) {
	pt1=strchr(copy,'\0');
	pt1 -= strlen(sufx[ender]);
	*pt1='\0';
	strcat(copy,addr[ender]);
    }
    return(copy);
}

static char *exc_lookup(word, dbase)
char *word, *dbase;
{
    register char *line, *beglp, *endlp;
    FILE *fp;
    char *env;
    char searchdir[256], fname[256];
    
    /* compute searchdir based on environment variable if present */
    
    strcpy(searchdir,((env = getenv("WNSEARCHDIR")) ? env : DEFAULTPATH));
#ifdef PC
    sprintf(fname,"%s\\%s",searchdir,dbase);
#else
    sprintf(fname,"%s/%s",searchdir,dbase);
#endif
    
    if ((fp = fopen(fname, "r")) == NULL) {
	fprintf(stderr, "Error in opening exception file(%s)\n\n",fname);
	return("");
    }
    
    if ((line = bin_search(word, fp)) != NULL) {
	beglp=strchr(line,' ');
	beglp++;
	endlp=strchr(beglp,'\n');
	if (*(endlp - 1) == ' ')
	    endlp--;
	*endlp='\0';
	return(beglp);
    } else
	return("");
}

char *morphword(word,pos)
char *word;
int pos;
{
    int offset, cnt;
    int i;
    static char retval[WORDSIZE];
    
    sprintf(retval,"");

    offset = offsets[pos];
    cnt = cnts[pos];

    /* first look for word on exception list */

    strcpy(retval, exc_lookup(word, exc_fnames[pos]));
    if(strcmp(retval,"") && is_defined(retval, pos))	/* found it */
	return(retval);

    if (pos == NOUN) {
	if(strend(word, "ss") || /* check for noun ending with 'ss' */
	   (strlen(word) <= 2)) { /* word is too short to morph */
    	    strcpy(retval, word);
	    return(retval);
	}
    }

    for(i = 0; i < cnt; i++) {
	strcpy(retval, wordbase(word, (i + offset)));
	if(strcmp(retval, word)) {
	    if(is_defined(retval, pos))
		return(retval);
	}
    }
    strcpy(retval, word);
    return(retval);
}

char *morphstr(origstr, pos)
char *origstr;
int pos;
{
    static char searchstr[WORDSIZE];
    char word[WORDSIZE], str[WORDSIZE];
    int cnt, st_idx = 0, end_idx;
    
    strtolower(strsubst(strcpy(str, origstr), ' ', '_'));
    searchstr[0] = '\0';
    cnt = cntwords(str);
    while (--cnt) {
	end_idx = strchr(str + st_idx, '_') - str;
	strncpy(word, str + st_idx, end_idx - st_idx);
	word[end_idx - st_idx] = '\0';
	strcat(searchstr, morphword(word, pos));
	strcat(searchstr, "_");
	st_idx = end_idx + 1;
    }
    
    /* Pick up last word in string */
    
    strcat(searchstr, morphword(strcpy(word, str + st_idx), pos));
    
    return(searchstr);
}

/*
  Revision log:
  
  $Log:	morph.c,v $
 * Revision 1.10  92/04/07  14:03:26  wn
 * added test for noun <= 2 characters
 * 
 * Revision 1.9  92/03/25  10:04:45  wn
 * *** empty log message ***
 * 
 * Revision 1.8  92/03/19  11:47:57  wn
 * lots of cleanup, removed binary search to use one in libarary
 * 
  * Revision 1.7  92/02/25  14:38:57  wn
  * changed index to strchr and added DEFAULTPATH to port to PC
  * 
  * Revision 1.6  92/02/21  16:29:00  wn
  * moved cntwords to util.c
  * 
  * Revision 1.5  92/02/12  10:35:15  wn
  * added static to internal funtions
  * 
  * Revision 1.4  92/02/05  11:08:25  wn
  * added morphstr() and cntwords()
  * changed morphit() to morphword() and changed 2nd arg to int
  * 
  * Revision 1.3  91/12/13  15:21:08  wn
  * changed 2nd arg to morphit to be single char instead of string
  * 
  * Revision 1.2  91/12/04  10:49:08  wn
  * don't check searchcode to see if word is already known and then
  * skip morph.  always morph word and return word, even if the same
  * 
  * Revision 1.1  91/09/25  15:39:47  wn
  * Initial revision
  * 
  */
