/*
  
  search.c - WordNet search algorithm
  
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <assert.h>
#include "wn.h"
#include "wnconsts.h"
#include "wnglobals.h"

static char *Id = "$Id: search.c,v 1.35 92/04/07 16:37:51 wn Exp $";
/* For adjectives, indictes synset type */

#define DONT_KNOW	0
#define DIRECT_ANT	1	/* direct antonyms (cluster head) */
#define INDIRECT_ANT	2	/* indrect antonyms (similar) */
#define PERTAINYM	3	/* no antonyms or similars (pertainyms) */

#define LINEBUF		10240	/* 10K buffer to read in index & data files */
#define SMLINEBUF	1024	/* small buffer for output lines */

/* forward declarations */

static void printsynset(), printadjsynset(), printbuffer(), printsense(),
    readfnames();
static char *printant();

int SenseCount;			/* number of senses word has */
int OutSenseCount;		/* number of senses search found */
int dflag = 1;			/* if set, print definitional glosses */
char *(*wn_filenames[4])[] = { NULL, NULL, NULL, NULL }; 

static int prflag, sense, prsense;
static char *searchbuffer = NULL;
static char tmpbuf[1024];	/* general purpose printing buffer */
static char wdbuf[256];		/* general purpose word buffer */
static int adj_marker;
static char *markers[] = {
    "",
    "(prenominal)",
    "(postnominal)",
    "(predicate)",
};

static char *freqcats[] = {
    "extremely rare","very rare","rare","uncommon","common",
    "familiar","very familiar","extremely familiar"
    };

/*
 * Tracked-malloc routine.
 */

int blockCount =0;

typedef struct block {
    char *address;
    struct block *next;
} Block;

Block *thead = NULL;
Block *ttail = NULL;

static char *tmalloc(size)
unsigned long size;
{
    /* Create a block and allocate the memory. */
    Block *block = (Block *)malloc(sizeof(struct block));
    block->address = malloc(size);
    block->next = NULL;
    ++blockCount;
    
    /* Link the block into the list. */
    if (thead == NULL) {
	thead = block;
	ttail = block;
    }
    else {
	ttail->next = block;
	ttail = block;
    }
    
    return block->address;
}

void freeAll()
{
    Block *block = thead;
    Block *next;
    
    /* Walk the list of blocks and free all memories. */
    while (block != NULL) {
	next = block->next;
	free(block->address);
	free(block);
	
	--blockCount;
	block = next;
    }
    thead = NULL;
    ttail = NULL;
    free(searchbuffer);
    searchbuffer = NULL;
}

/*
 * Given a word and a part of speech marker ADJ,NOUN,VERB
 * returns an index entry structure IndexPtr.
 */
IndexPtr index_lookup(word,dbase)
char *word;
int  dbase;
{
    register char *ptrtok;
    register IndexPtr idx = NULL;
    register int j,i;
    char *env;
    FILE *fp;
    char searchdir[256], fname[256];
    char *line;
    
    strcpy(searchdir,((env=getenv("WNSEARCHDIR")) ? env : DEFAULTPATH));
#ifdef PC
    sprintf(fname,"%s\\%s.idx",searchdir,partnames[dbase]);
#else
    sprintf(fname,"%s/index.%s",searchdir,partnames[dbase]);
#endif
    
    if((fp=fopen(fname,"r")) == NULL) {
	fprintf(stderr, "Error: can't open indexfile(%s)\n",fname);
	exit(-1);
    }
    
    if ((line = bin_search(word, fp)) != NULL) {
        idx = (IndexPtr)tmalloc(sizeof(Index));
        assert(idx);
	
        idx->wd='\0';
        idx->pos='\0';
        idx->off_cnt=0;
        idx->sense_cnt=0;
        idx->offset='\0';
        idx->ptruse_cnt=0;
        idx->ptruse='\0';
	
	/* get the word */
	ptrtok=strtok(line," \n");
	
	idx->wd=(char *) tmalloc(strlen(ptrtok) + 1);
	assert(idx->wd);
	strcpy(idx->wd, ptrtok);
	
	/* get the part of speech */
	ptrtok=strtok(NULL," \n");
	idx->pos=(char *) tmalloc(strlen(ptrtok) + 1);
	assert(idx->pos);
	strcpy(idx->pos, ptrtok);
	
	/* get the collins count */
	ptrtok=strtok(NULL," \n");
	idx->sense_cnt = atoi(ptrtok);
	
	/* get the number of pointers types */
	ptrtok=strtok(NULL," \n");
	idx->ptruse_cnt = atoi(ptrtok);
	
	idx->ptruse = (int *) tmalloc(idx->ptruse_cnt * (sizeof(int)));
	assert(idx->ptruse);
	
	/* get the pointers types */
	for(j=0;j < idx->ptruse_cnt; j++) {
	    ptrtok=strtok(NULL," \n");
	    for(i=1;i<=MAXSEARCH;i++) {
		if(!strcmp(ptrtok,ptrtyp[i]))
		    idx->ptruse[j]=i;
	    }
	}
	
	/* get the number of offsets */
	ptrtok=strtok(NULL," \n");
	idx->off_cnt = atoi(ptrtok);
	
	/* make space for the offsets */
	idx->offset = (long *) tmalloc(idx->off_cnt * (sizeof(long)));
	assert(idx->offset);
	
	/* get the offsets */
	for(j=0;j<idx->off_cnt;j++) {
	    ptrtok=strtok(NULL," \n");
	    idx->offset[j] = atol(ptrtok);
	}
    }
    return(idx);
}

SynsetPtr read_synset(dbase,boffset,word)
int  dbase;
long boffset;
char *word;
{
    char searchdir[512], dataname[512];
    FILE *fp;
    static char line[LINEBUF];
    register char *ptrtok;
    char *tmpptr, *env;
    char too[3],tfrm[3];
    register int i,j;
    register SynsetPtr synptr;
    
    synptr = (SynsetPtr)tmalloc(sizeof(Synset));
    assert(synptr);
    
    synptr->hereiam=0;
    synptr->sstype=DONT_KNOW;
    synptr->fnum=0;
    synptr->pos='\0';
    synptr->wcount=0;
    synptr->words='\0';
    synptr->whichword=0;
    synptr->ptrcount=0;
    synptr->ptrtyp='\0';
    synptr->ptroff='\0';
    synptr->ppos = '\0';
    synptr->pto='\0';
    synptr->pfrm='\0';
    synptr->fcount=0;
    synptr->frmid='\0';
    synptr->frmto='\0';
    synptr->defn='\0';
    synptr->nextss = NULL;
    synptr->searchtype = -1;
    synptr->ptrlist = NULL;
    
    strcpy(searchdir,((env=getenv("WNSEARCHDIR")) ? env : DEFAULTPATH));
#ifdef PC
    sprintf(dataname,"%s\\%s.dat",searchdir,partnames[dbase]);
#else
    sprintf(dataname,"%s/data.%s",searchdir,partnames[dbase]);
#endif
    if((fp=fopen(dataname,"r")) == NULL) {
	fprintf(stderr, "Error: can't open datafile(%s)\n",dataname);
	exit(-1);
    }
    
    if (!wn_filenames[dbase])
	readfnames(fp, dbase);
    
    fseek(fp,boffset,0);
    fgets(line, LINEBUF, fp);
    fclose(fp);
    
    ptrtok=line;
    
    if (dbase==ADJ) {
	if (strchr(line,'!'))
	    synptr->sstype = DIRECT_ANT; /* cluster head */
	else if (strchr(line, '&'))
	    synptr->sstype = INDIRECT_ANT; /* similiar */
	else if (strchr(line, '\\'))
	    synptr->sstype = PERTAINYM; /* pertainym */
    }
    
    /* looking at offset */
    ptrtok=strtok(line," \n");
    synptr->hereiam = atol(ptrtok);
    
    /* looking at FNUM */
    ptrtok=strtok(NULL," \n");
    synptr->fnum=atoi(ptrtok);
    
    /* looking at POS */
    ptrtok=strtok(NULL," \n");
    synptr->pos = (char *) tmalloc(strlen(ptrtok) + 1);
    assert(synptr->pos);
    strcpy(synptr->pos, ptrtok);
    
    /* looking at numwords */
    ptrtok=strtok(NULL," \n");
    synptr->wcount=strtol(ptrtok,NULL,16);
    
    synptr->words=(char **) tmalloc(synptr->wcount  * sizeof(char *));
    assert(synptr->words);
    synptr->wnsense = (int *) tmalloc(synptr->wcount * sizeof(int));
    assert(synptr->wnsense);
    
    for(i=0;i<synptr->wcount; i++) {
	ptrtok=strtok(NULL," \n");
	synptr->words[i]= (char *) tmalloc(strlen(ptrtok) + 1);
	assert(synptr->words[i]);
	strcpy(synptr->words[i], ptrtok);
	
	/* is this the word we're looking for? */
	
	if(!strcmp(word,strtolower(ptrtok))) /* strtolower does deadjify */
	    synptr->whichword = i+1;
	
	ptrtok = strtok(NULL, " \n"); 
	synptr->wnsense[i] = atoi(ptrtok);
    }
    
    /* get the pointer count */
    ptrtok=strtok(NULL," \n");
    synptr->ptrcount=atoi(ptrtok);
    
    /* alloc storage for the pointers */
    synptr->ptrtyp=(int *) tmalloc(synptr->ptrcount *sizeof(int));
    assert(synptr->ptrtyp);
    synptr->ptroff=(long *) tmalloc(synptr->ptrcount * sizeof(long));
    assert(synptr->ptroff);
    synptr->ppos = (int *) tmalloc(synptr->ptrcount * sizeof(int));
    assert(synptr->ppos);
    synptr->pto=(int *) tmalloc(synptr->ptrcount * sizeof(int));
    assert(synptr->pto);
    synptr->pfrm=(int *) tmalloc(synptr->ptrcount * sizeof(int));
    assert(synptr->pfrm);
    
    for(i=0;i<synptr->ptrcount;i++) {
	/* get the pointer type */
	ptrtok=strtok(NULL," \n");
	for(j=1;j<=MAXSEARCH;j++)
	    if(!strcmp(ptrtok,ptrtyp[j]))
		synptr->ptrtyp[i]=j;
	
	/* get the pointer offset */
	ptrtok=strtok(NULL," \n");
	synptr->ptroff[i] = atol(ptrtok);
	
	/* get the pointer part of speech */
	ptrtok = strtok(NULL, " \n");
	synptr->ppos[i] = getpos(ptrtok);
	
	/* get the lexp to/from restrictions */
	ptrtok=strtok(NULL," \n");
	
	tmpptr=ptrtok;
	strncpy(tfrm,tmpptr,2);
	tfrm[2]='\0';
	synptr->pfrm[i]=atoi(tfrm);
	
	tmpptr += 2;
	strncpy(too,tmpptr,2);
	too[2]='\0';
	synptr->pto[i]=atoi(too);
    }
    
    /* retireve optional information from verb synset */
    if(dbase == VERB) {
	ptrtok=strtok(NULL," \n");
	synptr->fcount=atoi(ptrtok);
	
	/* allocate frame storage */
	
	synptr->frmid=(int *) tmalloc(synptr->fcount * sizeof(int));  
	assert(synptr->frmid);
	synptr->frmto=(int *) tmalloc(synptr->fcount * sizeof(int));  
	assert(synptr->frmto);
	
	for(i=0;i<synptr->fcount;i++) {
	    /* skip the frame pointer (+) */
	    ptrtok=strtok(NULL," \n");
	    
	    ptrtok=strtok(NULL," \n");
	    synptr->frmid[i]=atoi(ptrtok);
	    
	    ptrtok=strtok(NULL," \n");
	    synptr->frmto[i]=atoi(ptrtok);
	}
    }
    
    /* get the optional definition */
    
    ptrtok=strtok(NULL," \n");
    if(ptrtok) {
	ptrtok=strtok(NULL," \n");
	sprintf(tmpbuf,"");
	while(ptrtok != NULL) {
	    strcat(tmpbuf,ptrtok);
	    ptrtok=strtok(NULL," \n");
	    if(ptrtok)
		strcat(tmpbuf," ");
	}
	synptr->defn=(char *) tmalloc(strlen(tmpbuf) + 10);
	assert(synptr->defn);
	sprintf(synptr->defn,"(%s)",tmpbuf);
    }
    
    return(synptr);
}

static void traceptrs(synptr,ptrtyp,dbase,depth)
SynsetPtr synptr;
int ptrtyp;
int dbase;
int depth;
{
    register int i,j;
    register SynsetPtr cursyn;
    char prefix[100];
    
    for(i=0;i<synptr->ptrcount;i++) {
	if((synptr->ptrtyp[i] == ptrtyp) &&
	   ((synptr->pfrm[i] == 0) ||
	    (synptr->pfrm[i] == synptr->whichword))) {
	    
	    if(!prflag) {
		printsense();
		printsynset("",synptr,"\n",DEFON);
		prflag = 1;
	    }
	    
	    for(j=0;j<depth;j++)
		printbuffer("    ");
	    
	    printbuffer("   ");
	    
	    if(!depth)
		printbuffer("       ");
	    
	    switch(ptrtyp) {
	    case SEEALSOPTR:
		sprintf(prefix, "See Also -> ");
		break;
	    case PERTPTR:
		sprintf(prefix, "Pertains to (%s) -> ",
			partnames[synptr->ppos[i]]);
		break;
	    case HASMEMBERPTR:
		sprintf(prefix, "   HAS MEMBER: ");
		break;
	    case HASSTUFFPTR:
		sprintf(prefix, "   HAS SUBSTANCE: ");
		break;
	    case HASPARTPTR:
		sprintf(prefix, "   HAS PART: ");
		break;
	    case ISMEMBERPTR:
		sprintf(prefix, "   MEMBER OF: ");
		break;
	    case ISSTUFFPTR:
		sprintf(prefix, "   SUBSTANCE OF: ");
		break;
	    case ISPARTPTR:
		sprintf(prefix, "   PART OF: ");
		break;
	    default:
		sprintf(prefix, "=> ");
		break;
	    }
	    
	    cursyn=read_synset(synptr->ppos[i], synptr->ptroff[i], "");
	    printsynset(prefix, cursyn, "\n", DEFON);
	    
	    if(depth) {
		depth = depthcheck(depth, cursyn);
		traceptrs(cursyn,ptrtyp,dbase,(depth+1));
	    }	
	}
    }
}

SynsetPtr traceptrs_ds(synptr,ptrtyp,dbase,depth)
SynsetPtr synptr;
int ptrtyp;
int dbase;
int depth;
{
    register int i;
    register SynsetPtr cursyn, synlist = NULL, lastsyn = NULL;
    
    for(i=0;i<synptr->ptrcount;i++) {
	if((synptr->ptrtyp[i] == ptrtyp) &&
	   ((synptr->pfrm[i] == 0) ||
	    (synptr->pfrm[i] == synptr->whichword))) {
	    
	    cursyn=read_synset(synptr->ppos[i], synptr->ptroff[i], "");
	    cursyn->searchtype = ptrtyp;
	    
	    if (lastsyn)
		lastsyn->nextss = cursyn;
	    if (!synlist)
		synlist = cursyn;
	    lastsyn = cursyn;
	    
	    if(depth) {
		depth = depthcheck(depth, cursyn);
		cursyn->ptrlist = traceptrs_ds(cursyn,ptrtyp,dbase,(depth+1));
	    }	
	}
    }
    return(synlist);
}

static void tracecoords(synptr,ptrtyp,dbase,depth)
SynsetPtr synptr;
int ptrtyp;
int dbase;
int depth;
{
    register int i,j;
    register SynsetPtr cursyn;
    
    for(i=0;i<synptr->ptrcount;i++) {
	if((synptr->ptrtyp[i] == HYPERPTR) &&
	   ((synptr->pfrm[i] == 0) ||
	    (synptr->pfrm[i] == synptr->whichword))) {
	    
	    if(!prflag) {
		printsense();
		printsynset("",synptr,"\n",DEFON);
		prflag = 1;
	    }
	    
	    for(j=0;j<depth;j++)
		printbuffer("    ");
	    if(!depth)
		printbuffer("    ");
	    
	    cursyn=read_synset(dbase,synptr->ptroff[i],"");
	    printsynset("-> ",cursyn,"\n",DEFON);
	    
	    traceptrs(cursyn,ptrtyp,NOUN,depth);
	    
	    if(depth) {
		depth = depthcheck(depth, cursyn);
		tracecoords(cursyn,ptrtyp,dbase,(depth+1));
	    }	
	}
    }
}

/* Trace through the hypernym tree and print all MEMBER, STUFF
   and PART info. */

static void traceinherit(synptr,ptrbase,dbase,depth)
SynsetPtr synptr;
int ptrbase;
int dbase;
int depth;
{
    register int i,j;
    register SynsetPtr cursyn;
    
    for(i=0;i<synptr->ptrcount;i++) {
	if((synptr->ptrtyp[i] == HYPERPTR) &&
	   ((synptr->pfrm[i] == 0) ||
	    (synptr->pfrm[i] == synptr->whichword))) {
	    
	    if(!prflag) {
		printsense();
		printsynset("",synptr,"\n",DEFON);
		prflag = 1;
	    }
	    
	    for(j=0;j<depth;j++)
		printbuffer("    ");
	    if(!depth)
		printbuffer("\n    ");
	    
	    cursyn=read_synset(dbase,synptr->ptroff[i],"");
	    printsynset("=> ",cursyn,"\n",DEFON);
	    
	    traceptrs(cursyn,ptrbase,NOUN,depth);
	    traceptrs(cursyn,ptrbase + 1,NOUN,depth);
	    traceptrs(cursyn,ptrbase + 2,NOUN,depth);
	    
	    if(depth) {
		depth = depthcheck(depth, cursyn);
		traceinherit(cursyn,ptrbase,dbase,(depth+1));
	    }	
	}
    }
}

static void traceadjant(synptr)
SynsetPtr synptr;
{
    register int i, j;
    int anttype = DIRECT_ANT;
    register SynsetPtr simptr, antptr;
    static char similar[] = "        => ";
    
    /* This search is only applicable for ADJ synsets which have
       either direct or indirect antonyms (not valid for pertainyms). */
    
    if (synptr->sstype == DIRECT_ANT || synptr->sstype == INDIRECT_ANT) {
	printsense();
	printsynset("", synptr, "\n\n", DEFON);
	
	/* if indirect, get cluster head */
	
	if(synptr->sstype == INDIRECT_ANT) {
	    anttype = INDIRECT_ANT;
	    i = 0;
	    while (synptr->ptrtyp[i] != SIMPTR) i++;
	    synptr = read_synset(ADJ, synptr->ptroff[i], "");
	}
	
	/* find antonyms - if direct, make sure that the antonym
	   ptr we're looking at is from this word */
	
	for (i = 0; i < synptr->ptrcount; i++) {
	    if (synptr->ptrtyp[i] == ANTPTR &&
		((anttype == DIRECT_ANT &&
		  synptr->pfrm[i] == synptr->whichword) ||
		 (anttype == INDIRECT_ANT))) {
		
		/* read the antonym's synset and print it.  if a
		   direct antonym, print it's satellites. */
		
		antptr = read_synset(ADJ, synptr->ptroff[i], "");
		printadjsynset(antptr, "\n", anttype, DEFON);

		if (anttype == DIRECT_ANT) {
		    for(j = 0; j < antptr->ptrcount; j++) {
			if(antptr->ptrtyp[j] == SIMPTR) {
			    simptr = read_synset(ADJ, antptr->ptroff[j], "");
			    printsynset(similar, simptr, "\n", DEFON);
			}
		    }
		}	
	    }
	}
    }
}

static void partsall(synptr, ptrtyp)
SynsetPtr synptr;
int ptrtyp;
{
    int ptrbase;
    int cnt, total = 0;
    
    ptrbase = (ptrtyp == HMERONYM) ? HASMEMBERPTR : ISMEMBERPTR;
    
    /* First, print out the MEMBER, STUFF, PART info for this synset */
    
    if (cnt = CntTreePtrs(synptr, ptrbase, NOUN, 1, 0)) {
	traceptrs(synptr, ptrbase, NOUN, 1);
	total += cnt;
    }
    if (cnt = CntTreePtrs(synptr, ptrbase + 1, NOUN, 1, 0)) {
	traceptrs(synptr, ptrbase + 1, NOUN, 1);
	total += cnt;
    }
    if (cnt = CntTreePtrs(synptr, ptrbase + 2, NOUN, 1, 0)) {
	traceptrs(synptr, ptrbase + 2, NOUN, 1);
	total += cnt;
    }
    
#ifdef FOOP
    if (total)	
#endif
    /* Print out MEMBER, STUFF, PART info for hypernyms on
       HMERONYM search only */

    if (total && ptrtyp == HMERONYM) 
	traceinherit(synptr, ptrbase, NOUN, 1);
}

static void printframe(synptr)
SynsetPtr synptr;
{
    register int i;
    
    printsense();
    printsynset("",synptr,"\n",DEFON);
    
    for(i = 0; i < synptr->fcount; i++) {
	if ((synptr->frmto[i] == synptr->whichword) ||
	    (synptr->frmto[i] == 0)) {
	    if (synptr->frmto[i] == synptr->whichword)
		printbuffer("\t=> ");
	    else
		printbuffer("\t*> ");
	    printbuffer(frametext[synptr->frmid[i]]);
	    printbuffer("\n");
	}
    }
}

static void freq_word(word,dbase)
char *word;
int dbase;
{
    register IndexPtr index;
    register int familiar=0;
    register int cnt;
    static char *a_an[] = { "", "a noun", "a verb", "an adjective" };
    
    if((index=index_lookup(word,dbase)) != NULL) {
	cnt = index->sense_cnt;
	
	if (cnt == 0) familiar = 0;
	if (cnt == 1) familiar = 1;
	if (cnt == 2) familiar = 2;
	if (cnt >= 3 && cnt <= 4) familiar = 3;
	if (cnt >= 5 && cnt <= 8) familiar = 4;
	if (cnt >= 9 && cnt <= 16) familiar = 5;
	if (cnt >= 17 && cnt <= 32) familiar = 6;
	if (cnt > 32 ) familiar = 7;

	sprintf(tmpbuf,
		"\n%s used as %s is %s (polysemy count = %d)\n",
		word, a_an[dbase], freqcats[familiar], cnt);
	printbuffer(tmpbuf);
    }
}

/* Determine if any of the synsets that this word is in have inherited
   meronyms or holonyms. */

static int CntHoloMero(index, ptrtyp)
IndexPtr index;
int ptrtyp;
{
    register int i;
    register SynsetPtr synset;
    register int total=0;
    int ptrbase;
    
    ptrbase = (ptrtyp == HMERONYM) ? HASMEMBERPTR : ISMEMBERPTR;
    
    for(i = 0; i < index->off_cnt; i++) {
	synset = read_synset(NOUN, index->offset[i], "");
	total += CntTreePtrs(synset, ptrbase, NOUN, 1, 0);
	total += CntTreePtrs(synset, ptrbase + 1, NOUN, 1, 0);
	total += CntTreePtrs(synset, ptrbase + 2, NOUN, 1, 0);
    }
    return(total);
}

/* Trace through the hypernym tree and count all occurrences of
   the pointer type passed.  Used to find inherited mero/holonyms. */

static int CntTreePtrs(synptr,ptrtyp,dbase,depth,numparts)
SynsetPtr synptr;
int ptrtyp;
int dbase;
int depth;
int numparts;
{
    register int i,j;
    static int count;
    register SynsetPtr cursyn;
    
    count = numparts;
    
    for(i = 0; i < synptr->ptrcount; i++)
	if(synptr->ptrtyp[i] == ptrtyp)
	    count++;
    
    for(i = 0; i < synptr->ptrcount; i++) {
	if(synptr->ptrtyp[i] == HYPERPTR) { 
	    cursyn=read_synset(dbase,synptr->ptroff[i],"");
	    for(j = 0; j < cursyn->ptrcount; j++) {
		if(cursyn->ptrtyp[j] == ptrtyp)
		    count++;
	    }
	    if(depth) {
		depth = depthcheck(depth, cursyn);
		CntTreePtrs(cursyn,ptrtyp,dbase,depth+1,count);
	    }	
	}
    }
    return(count);
}

char *findtheinfo(word,dbase,ptrtyp)
char *word;
int dbase;
int ptrtyp;
{
    IndexPtr idx;
    SynsetPtr cursyn;
    int depth=0;
    
    if(ptrtyp < 0) {
	ptrtyp *= -1;
	depth=1;
    }
    
    /* initialize searchbuffer */
    
    if (!searchbuffer) {
	searchbuffer = (char *) malloc (FINDBUFSIZE * sizeof(char));
	assert(searchbuffer);
    }
    
    sprintf(searchbuffer,"");
    
    if(ptrtyp == FREQ) {
	freq_word(word,dbase);
    } else {
	if ((idx=index_lookup(word,dbase)) != NULL) {

	    OutSenseCount = 0;	/* keep track of # senses printed */
	    SenseCount = idx->off_cnt; /* number of senses word has */
	    
	    sprintf(tmpbuf,
		    "                                                                         \n");
	    printbuffer(tmpbuf);
	    
	    /* Go through all of the searchword's senses in the
	       database and perform the search requested. */
	    
	    for(sense = 0; sense < idx->off_cnt; sense++) {
		prflag=0;
		cursyn=read_synset(dbase,idx->offset[sense],idx->wd);
		
		switch(ptrtyp) {
		case ANTPTR:
		    if(dbase == ADJ)
			traceadjant(cursyn);
		    else
			traceptrs(cursyn,ANTPTR,dbase,depth);
		    break;
		    
		case COORDS:
		    tracecoords(cursyn,HYPOPTR,dbase,depth);
		    break;
		    
		case FRAMES:
		    printframe(cursyn);
		    break;
		    
		case MERONYM:
#ifdef FOOP
		    traceptrs(cursyn,MERONYM,dbase,depth);
#endif
		    traceptrs(cursyn,HASMEMBERPTR,dbase,depth);
		    traceptrs(cursyn,HASSTUFFPTR,dbase,depth);
		    traceptrs(cursyn,HASPARTPTR,dbase,depth);
		    break;
		    
		case HOLONYM:
#ifdef FOOP
		    traceptrs(cursyn,HOLONYM,dbase,depth);
#endif
		    traceptrs(cursyn,ISMEMBERPTR,dbase,depth);
		    traceptrs(cursyn,ISSTUFFPTR,dbase,depth);
		    traceptrs(cursyn,ISPARTPTR,dbase,depth);
		    break;
		    
		case HMERONYM:
		    partsall(cursyn, HMERONYM);
		    break;
		    
		case HHOLONYM:
		    partsall(cursyn, HHOLONYM);
		    break;
		    
		case SIMPTR:
		case SYNS:
		case HYPERPTR:
		    printsense();
		    printsynset("",cursyn,"\n",DEFON);
		    prflag = 1;
		    
		    traceptrs(cursyn,ptrtyp,dbase,depth);
		    
		    if (dbase == ADJ)
			traceptrs(cursyn, PERTPTR, dbase, depth);
		    
		    traceptrs(cursyn,SEEALSOPTR,dbase,depth);
		    break;
		    
		default:
		    traceptrs(cursyn,ptrtyp,dbase,depth);
		    break;
		    
		}
	    }      
	    
	    /* patch in the number of senses output */
	    
	    printbuffer("\n");
	    if (OutSenseCount == idx->off_cnt) {
		if (OutSenseCount == 1)
		    sprintf(tmpbuf, "1 sense of %s", word);
		else
		    sprintf(tmpbuf, "%d senses of %s", OutSenseCount, word);
	    } else
		sprintf(tmpbuf, "%d of %d senses of %s",
			OutSenseCount, idx->off_cnt, word);
	    strncpy(searchbuffer, tmpbuf, strlen(tmpbuf));
	}
    }
    return(strsubst(searchbuffer,'_',' '));
}

SynsetPtr findtheinfo_ds(word,dbase,ptrtyp)
char *word;
int dbase;
int ptrtyp;
{
    IndexPtr idx;
    SynsetPtr cursyn;
    SynsetPtr synlist = NULL, lastsyn = NULL;
    int depth=0;
    
    if(ptrtyp < 0) {
	ptrtyp *= -1;
	depth=1;
    }
    
    if ((idx=index_lookup(word,dbase)) != NULL) {
	
	SenseCount = idx->off_cnt; /* number of senses word has */
	
	/* Go through all of the searchword's senses in the
	   database and perform the search requested. */
	
	for(sense = 0; sense < idx->off_cnt; sense++) {
	    cursyn=read_synset(dbase,idx->offset[sense],idx->wd);
	    if (lastsyn)
		lastsyn->nextss = cursyn;
	    if (!synlist)
		synlist = cursyn;
	    
	    cursyn->searchtype = ptrtyp;
	    cursyn->ptrlist = traceptrs_ds(cursyn,ptrtyp,dbase,depth);
	    
	    lastsyn = cursyn;
	}      
	
    }
    return(synlist);
}

unsigned long is_defined(word,dbase)
char *word;
int  dbase;
{
    IndexPtr index;
    int i;
    unsigned long retval=0;
    
    if((index = index_lookup(word,dbase)) != NULL) {
	
	/* set bits that must be true for all words */
	
	retval = bit(SIMPTR) | bit(FREQ) | bit(SYNS);
	
	if (dbase == NOUN) {
	    if (CntHoloMero(index, HMERONYM))
		retval |= bit(HMERONYM);
	    if (CntHoloMero(index, HHOLONYM))
		retval |= bit(HHOLONYM);
	}
	
	for(i = 0; i < index->ptruse_cnt; i++) {
	    
	    retval |= bit(index->ptruse[i]); /* set bit for this ptr */
	    
	    /* set generic HOLONYM and/or MERONYM bit if necessary */
	    
	    if(index->ptruse[i] >= ISMEMBERPTR &&
	       index->ptruse[i] <= ISPARTPTR)
		retval |= bit(HOLONYM);
	    else if(index->ptruse[i] >= HASMEMBERPTR &&
		    index->ptruse[i] <= HASPARTPTR)
		retval |= bit(MERONYM);
	    else if (index->ptruse[i] == SIMPTR && dbase == ADJ)
		retval |= bit(ANTPTR);
	}
    }
    return(retval);
}

/* Strip off () enclosed comments from a word */

static char *deadjify(word)
char *word;
{
    char *y;
    
    adj_marker = UNKNOWN_MARKER; /* default if not adj or unknown */
    
    y=word;
    while(*y) {
	if(*y == '(') {
	    if (!strncmp(y, "(a)", 3))
		adj_marker = ATTRIBUTIVE;
	    else if (!strncmp(y, "(ip)", 4))
		adj_marker = IMMED_POSTNOMINAL;
	    else if (!strncmp(y, "(p)", 3))
		adj_marker = PREDICATIVE;
	    *y='\0';
	} else 
	    y++;
    }
    return(word);
}

static void printadjsynset(synptr, tail, anttype, definition)
char *tail;
SynsetPtr synptr;
int anttype, definition;
{
    register int i, wdcnt;
    char *str;
    char tbuf[SMLINEBUF];

    if (anttype == DIRECT_ANT) {
	printsynset("", synptr, tail, definition);
	return;
    }
    
    sprintf(tbuf,"");
    
    if((str = getenv("WNDEBUG")) && *str != '\0') {
	sprintf(tbuf,"<%s>",
		(*wn_filenames[getpos(synptr->pos)])[synptr->fnum]);
	prsense = 1;
    } else
	prsense = 0;
    
    /* print anotnyms from cluster head (of indirect ant) */
    
    strcat(tbuf, "INDIRECT (VIA ");
    for(i=0, wdcnt = synptr->wcount; i < wdcnt; i++) {
	str = printant(ADJ, synptr, i + 1, "%s", ", ");
	if (*str)
	    strcat(tbuf, str);
    }
    strcat(tbuf, ") -> ");
    
    /* now print synonyms from cluster head (of indirect ant) */
    
    for (i = 0, wdcnt = synptr->wcount; i < wdcnt; i++) {
	strcpy(wdbuf, synptr->words[i]);
	strcat(tbuf, deadjify(wdbuf));

	/* if debug is set, add WordNet sense number if non-0 */

	if (prsense && (synptr->wnsense[i] != 0))
	    sprintf(tbuf + strlen(tbuf), "%d", synptr->wnsense[i]);
#ifdef FOOP
	/* don't print adj marker in this case */
	if (adj_marker != UNKNOWN_MARKER) /* set in deadjify() */
	    strcat(tbuf, markers[adj_marker]);
#endif
	if (i < wdcnt - 1)
	    strcat(tbuf,", ");
    }
    
    if(dflag && synptr->defn && definition) {
	strcat(tbuf," -- ");
	strcat(tbuf,synptr->defn);
    }
    
    strcat(tbuf,tail);
    printbuffer(tbuf);
}

static void printsynset(head, synptr, tail, definition)
char *head, *tail;
SynsetPtr synptr;
int definition;
{
    register int i, wdcnt;
    char *str;
    char tbuf[SMLINEBUF];
    static char vs[] = " (vs. %s)";

    sprintf(tbuf,"");
    
    if((str = getenv("WNDEBUG")) && *str != '\0') {
	sprintf(tbuf,"<%s>",
		(*wn_filenames[getpos(synptr->pos)])[synptr->fnum]);
	prsense = 1;
    } else
	prsense = 0;
    
    strcat(tbuf,head);
    
    /* copy the word (since deadjify() changes original string),
       deadjify() the copy and append to buffer.  then append
       adjective marker string if there is one. */
    
    for(i=0, wdcnt = synptr->wcount; i < wdcnt; i++) {
	strcpy(wdbuf, synptr->words[i]);
	strcat(tbuf, deadjify(wdbuf));

	/* if debug is set, add WordNet sense number if non-0 */

	if (prsense && (synptr->wnsense[i] != 0))
	    sprintf(tbuf + strlen(tbuf), "%d", synptr->wnsense[i]);

	/* append adjective marker if present */

	if (getpos(synptr->pos) == ADJ) {
	    if (adj_marker != UNKNOWN_MARKER) /* set in deadjify() */
		strcat(tbuf, markers[adj_marker]);
	    strcat(tbuf,
		   printant(getpos(synptr->pos), synptr, i + 1, vs, ""));
	}
	if (i < wdcnt - 1)
	    strcat(tbuf,", ");
    }
    
    if(dflag && synptr->defn && definition) {
	strcat(tbuf," -- ");
	strcat(tbuf,synptr->defn);
    }
    
    strcat(tbuf,tail);
    printbuffer(tbuf);
}

static char *printant(dbase, synptr, wdnum, template, tail)
int dbase;
SynsetPtr synptr;
int wdnum;
char *template, *tail;
{
    register int i, j, wdoff;
    register SynsetPtr psynptr;
    char tbuf[256];
    static char retbuf[1024];
    int first = 1;
    
    retbuf[0] = '\0';

    /* Go through all the pointers looking for anotnyms from the word
       indicated by wdnum.  When found, print all the antonym's
       antonym pointers which point back to wdnum. */

    for (i = 0; i < synptr->ptrcount; i++) {
	if (synptr->ptrtyp[i] == ANTPTR && synptr->pfrm[i] == wdnum) {
	    psynptr = read_synset(dbase, synptr->ptroff[i], "");
	    for (j = 0; j < psynptr->ptrcount; j++) {
		if (psynptr->ptrtyp[j] == ANTPTR &&
		    psynptr->pto[j] == wdnum &&
		    psynptr->ptroff[j] == synptr->hereiam) {
		    
		    wdoff = psynptr->pfrm[j] - 1;
		    strcpy(wdbuf, psynptr->words[wdoff]);
		    if (prsense && (psynptr->wnsense[wdoff] != 0))
			sprintf(tbuf, "%s%d", deadjify(wdbuf),
				psynptr->wnsense[wdoff]);
		    else
			sprintf(tbuf, "%s", deadjify(wdbuf));
		    if (!first)
			strcat(retbuf, tail);
		    else
			first = 0;
		    sprintf(retbuf + strlen(retbuf), template, tbuf);
		}
	    }
	}
    }
    return(retbuf);
}

static void printbuffer(string)
char *string;
{
    int delta;
    static int findbufsize = FINDBUFSIZE; /* init to default size */
    
    delta = (findbufsize - strlen(searchbuffer)) - strlen(string);
    if(delta <= 0) {
	delta = delta * (-1);
	findbufsize += (delta + FINDBUFSIZE);
	searchbuffer = realloc(searchbuffer,findbufsize);
	assert(searchbuffer);
    }
    strcat(searchbuffer,string);
}

static void printsense()
{
    sprintf(tmpbuf,"\nSense %d\n", sense + 1);
    printbuffer(tmpbuf);
    OutSenseCount++;
}

static int getpos(s)
char *s;
{
    if (!strncmp(s, "n", 1))
	return(NOUN);
    else if (!strncmp(s, "a", 1) || !strncmp(s, "s", 1))
	return(ADJ);
    else if (!strncmp(s, "v", 1))
	return(VERB);
    else {
	fprintf(stderr, "Error: unknown part of speech\n");
	exit(-1);
    }
}

static void readfnames(fp, dbase)
FILE *fp;
int dbase;
{
    char line[SMLINEBUF];
    char *tmpptrs[100];		/* place to hold ptrs to fnames */
    char **p;
    int i = 0, len;
    
    while (1) {			/* skip license and copyright notice */
	fgets(line, SMLINEBUF, fp);
	if (line[0] != ' ')
	    break;
    }
    
    while (line[0] != '0') {
	len = 0;
	while (line[len++] != ' ');
	tmpptrs[i] = malloc(len);
	assert(tmpptrs[i]);
	strncpy(tmpptrs[i], line, len - 1);
	tmpptrs[i][len - 1] = '\0';	/* append null */
	i++;
	fgets(line, SMLINEBUF, fp);
    }
    
    p = (char **)malloc(i * sizeof(char *));
    assert(p);
    (char **)wn_filenames[dbase] = p;
    assert(wn_filenames[dbase]);
    
    while (i--)
	p[i] = tmpptrs[i];
}

static int depthcheck(depth, synptr)
int depth;
SynsetPtr synptr;
{
    if(depth>15) {
	if(depth == 16)
	    fprintf(stderr,"Error Cycle detected\n");
	fprintf(stderr,"   %s\n",synptr->words[0]);
	if(depth==20)
	    depth=-1;		/* reset to get one more trace then quit */
    }
    return(depth);
}

/*
  Revision log:
  
  $Log:	search.c,v $
 * Revision 1.35  92/04/07  16:37:51  wn
 * changes for PC - made buffers smaller and added SMLINEBUF
 * 
 * Revision 1.34  92/04/07  16:35:16  wn
 * removed usused variables
 * 
 * Revision 1.33  92/04/02  15:54:34  wn
 * change COORDS to use dbase rather than NOUN in traceptrs call
 * 
 * Revision 1.32  92/04/02  10:59:52  wn
 * made LINEBUF bigger
 * 
 * Revision 1.31  92/04/01  14:39:47  wn
 * changed familiary statement
 * 
 * Revision 1.30  92/03/26  14:45:30  wn
 * removed ISDEF
 * 
 * Revision 1.29  92/03/26  12:58:50  wn
 * *** empty log message ***
 * 
 * Revision 1.28  92/03/25  09:30:40  wn
 * fixed bug in printant
 * 
 * Revision 1.27  92/03/24  14:55:53  wn
 * removed BRACKET
 * 
 * Revision 1.26  92/03/24  14:08:10  wn
 * *** empty log message ***
 * 
  * Revision 1.25  92/03/12  11:45:11  wn
  * moved functions around, renamed some functions, added printadjant()
  * 
  * Revision 1.24  92/03/10  12:21:24  wn
  * added code to 'debug' output to print wn sense numbers
  * 
  * Revision 1.23  92/03/10  10:59:58  wn
  * lots of changes
  * 
  * Revision 1.22  92/02/25  14:39:27  wn
  * added DEFAULTPATH for port to pc.
  * changed sstype code to used #defines
  * 
  * Revision 1.21  92/02/21  16:29:28  wn
  * lots of cleanup
  * 
  * Revision 1.20  92/02/21  11:48:47  wn
  * added code to print holonyms and meronyms in more detail in 
  * output.
  * 
  * Revision 1.19  92/02/05  11:07:39  wn
  * changed some sprintf's to strcpy's and some other cleanups
  * 
  * Revision 1.18  92/01/22  15:56:53  wn
  * added delcarations for args to readfnames(O
  * 
  * Revision 1.17  92/01/08  16:57:07  wn
  * cleaned up readfnames()
  * changed synset * declations to synsetptr
  * 
  * Revision 1.16  92/01/03  11:52:43  wn
  * made strtolower and strsubst not static.
  * changed text of msgs printed in ISDEF search
  * 
  * Revision 1.15  91/12/18  11:54:52  wn
  * added global variables OutSenseCount and SenseCount so caller
  * can find out how many senses word has and how many were output
  * 
  * Revision 1.14  91/12/13  18:13:38  wn
  * fixed minor bug in readfnames
  * 
  * Revision 1.13  91/12/13  15:26:42  wn
  * clean up - 
  * changed name from find.c
  * made static functions if fn is used only within search.c
  * moved structures used only by search.c into here from wnconsts.h
  * 
  * Revision 1.12  91/12/11  10:14:12  wn
  * several changes - 
  * 1 - read input file names from data file rather than hard coding
  * strings into code
  * 2 - output number of senses at start of buffer
  * 3 - only allocate buffer 1st time search code is called
  * 
  * Revision 1.11  91/12/04  14:27:44  wn
  * fixed bug with sensecount
  * 
  * Revision 1.10  91/12/04  10:51:45  wn
  * changed msg text
  * 
  * Revision 1.9  91/12/04  10:49:50  wn
  * clean up msg reporting # of senses
  * 
  * Revision 1.8  91/12/04  10:08:46  wn
  * lots of cleanup
  * 
  * Revision 1.7  91/11/22  14:53:50  wn
  * added code for pertainyms and cleaned some things up
  * 
  * Revision 1.6  91/11/13  16:07:27  wn
  * increased line buffer size from 4k to 8k
  * general cleanup - removed unused variables, added register variables, etc.
  * added code to print out adjective marker strings if present.
  * 
  * Revision 1.5  91/10/30  10:52:46  wn
  * check WNDEBUG for null vs. non-null rather than for existance
  * 
  * Revision 1.4  91/10/23  16:34:29  wn
  * changed test in traceptrs to test pfrm instead of ptoo.
  * (see FOOP - old code commented out).
  * 
  * Revision 1.3  91/10/02  14:08:56  wn
  * removed test which was preventing sense from printing if no
  * hyponym
  * 
  * Revision 1.2  91/09/25  16:16:20  wn
  * attempt to fix bug with "leaf" (no hyponym).  removed test for ptr
  * to file number 0
  * 
  * Revision 1.1  91/09/25  15:39:26  wn
  * Initial revision
  * 
  */
