#include <stdio.h>
#include <string.h>
#include "wn.h"
#include "wnconsts.h"
#include "wnglobals.h"
#include "license.h"

static char *Id = "$Id: wn.c,v 1.23 92/04/07 16:39:59 wn Exp $";


static struct {
    char *opts;
    int optptrs;
    int optparts;
} options [] = {
    { "-synsa", SIMPTR,	ADJ },
    { "-antsa", ANTPTR,	ADJ },
    { "-perta", PERTPTR, ADJ },
    { "-famla", FREQ, ADJ },
    { "-antsn", ANTPTR,	NOUN },
    { "-hypen", -HYPERPTR, NOUN },
    { "-hypon", HYPOPTR, NOUN },
    { "-meron", MERONYM, NOUN },
    { "-holon", HOLONYM, NOUN },
    { "-treen", -HYPOPTR, NOUN },
    { "-famln", FREQ, NOUN },
    { "-synsn", HYPERPTR, NOUN },
    { "-subsn", HASSTUFFPTR, NOUN },
    { "-partn", HASPARTPTR, NOUN },
    { "-membn", HASMEMBERPTR, NOUN },
    { "-sprtn", ISPARTPTR, NOUN },
    { "-smemn", ISMEMBERPTR, NOUN },
    { "-ssubn", ISSTUFFPTR, NOUN },
    { "-hmern", -HMERONYM, NOUN },
    { "-hholn",	-HHOLONYM, NOUN },
    { "-coorn", COORDS, NOUN },
    { "-antsv", ANTPTR, VERB },
    { "-synsv", HYPERPTR, VERB },
    { "-famlv", FREQ, VERB },
    { "-hypev", -HYPERPTR, VERB },
    { "-treev", -HYPOPTR, VERB },
    { "-hypov", HYPOPTR, VERB },
    { "-entav", ENTAILPTR, VERB },
    { "-framv", FRAMES, VERB },
    { "-causv", CAUSETO, VERB },
    { "-coorv", COORDS, VERB },
    { NULL, 0, 0 }
};

struct {
    char *template;
    char *option;
    char *helpstr;
} searchstr[] = {
    { NULL, NULL, NULL },
    { "-ants%c",	"-ants{n|v|a}",	"\t\tAntonyms", },
    { "-hype%c",	"-hype{n|v}",	"\t\tHypernyms", },
    { "-hypo%c, -tree%c",	"-hypo{n|v}, -tree{n|v}",
	  "\tHyponyms & Hyponym Tree", },
    { "-enta%c",	"-entav",	"\t\tVerb Entailment", },
    { "-syns%c",	"-syns{n|v|a}",	"\t\tSynonyms", },
    { "-smem%c",	"-smemn",	"\t\tMember of Holonyms", },
    { "-ssub%c",	"-ssubn",	"\t\tSubstance of Holonyms", },
    { "-sprt%c",	"-sprtn",	"\t\tPart of Holonyms", },
    { "-memb%c",	"-membn",	"\t\tHas Member Meronyms", },
    { "-subs%c",	"-subsn",	"\t\tHas Substance Meronyms", },
    { "-part%c",	"-partn",	"\t\tHas Part Meronyms", },
    { "-mero%c",	"-meron",	"\t\tAll Meronyms", },
    { "-holo%c",	"-holon",	"\t\tAll Holonyms", },
    { "-caus%c",	"-causv",	"\t\tVerb Cause to", },
    { NULL, NULL, NULL },	/* verb caused by - not used */
    { NULL, NULL, NULL },	/* see also - no specific search */
    { "-pert%c",	"-perta",	"\t\tAdjective Pertainyms", },
    { NULL, NULL, NULL },	/* syns - taken care of w/SIMPTR */
    { "-faml%c",	"-faml{n|v|a}",	"\t\tFamiliarity & Polysemy Count", },
    { "-fram%c",	"-framv",	"\t\tVerb Frames", },
    { "-coor%c",	"-coor{n|v}",	"\t\tCoordinate Sisters", },
    { "-hmer%c",	"-hmern",	"\t\tHierarchical Meronyms", },
    { "-hhol%c", 	"-hholn",	"\t\tHierarchical Holonyms" },
};

main(argc,argv)
int argc;
char *argv[];
{
    int i,j=0;
    int found=0;
    int errcount = 0;
    unsigned long search;
    char *pbuf;
    char *morphword;
#ifdef LABONLY
    int wndebug;
#endif
    
    dflag = 0;			/* turn off default definitions */
    
    if(argc < 2) {
	fprintf(stdout,
		"\nusage: wn word [-d] [-l] -searchtype [-searchtype...]\n");
	fprintf(stdout,"\nusage: wn [-l]\n");
	fprintf(stdout, "\n\t-d (optional definition with search)\n");
	fprintf(stdout, "\t-l (print license and copyright notice)\n");
	fprintf(stdout,"\nsearchtype is one of the following:\n");
	for (i = 1; i <= MAXSEARCH; i++)
	    if (searchstr[i].option)
		fprintf(stdout, "\t%s%s\n",
			searchstr[i].option, searchstr[i].helpstr);
	exit(-1);
    }
    
    if(argc == 2) {
	if (!strcmp("-l", argv[1]))
	    printf("%s", license);
	else {
	    
	    /* Fix case of string and replace spaces with underscores in
	       multi word phrases. */
	    
	    strtolower(argv[1]);
	    strsubst(argv[1], ' ', '_');
	    
	    for (i = 1; i <= NUMPARTS; i++) {
		if ((search = is_defined(argv[1], i)) != 0)
		    printsearches(argv[1], i, search);
		else
		    printf("\nNo information available for %s %s\n",
			    partnames[i], argv[1]);

		morphword = morphstr(argv[1], i);

		if (strcmp(morphword, argv[1])) {
		    if ((search = is_defined(morphword, i)) != 0)
			printsearches(morphword, i, search);
		    else
			printf("\nNo information available for %s %s\n",
			       partnames[i], morphword);
		}
	    }
	    freeAll();
	}
	exit(0);
    }
    
    for(i=0;i<argc;i++) {
#ifdef LABONLY
	wndebug = (getenv("WNDEBUG") == NULL) ? 0 : 1;
#endif
	if(!strcmp("-d",argv[i]))
	    dflag++;
	else if (!strcmp("-l", argv[i]))
	    printf("%s", license);
#ifdef LABONLY
	else if (!strcmp("-dbg", argv[i]))
	    putenv("WNDEBUG=on");
#endif
    }
    
    strsubst(argv[1], ' ', '_');
    strtolower(argv[1]);
    
    /* skip the word */
    j++;
    
    while(argv[++j]) {
	if(strcmp("-d",argv[j]) && strcmp("-l", argv[j])
#ifdef LABONLY
	   && strcmp("-dbg", argv[j])
#endif
	   ) {
	    i=0;
	    found = 0;
	    
	    while(options[i].opts) {
		if(!strcmp(options[i].opts,argv[j])) {
		    found++;
		    pbuf = findtheinfo(argv[1],
				       options[i].optparts,
				       options[i].optptrs);
		    printf("%s",pbuf);
		    freeAll();
		    morphword = morphstr(argv[1], options[i].optparts);
		    if (strcmp(morphword, argv[1])) {
			pbuf = findtheinfo(morphword,
					   options[i].optparts,
					   options[i].optptrs);
			printf("%s",pbuf);
		        freeAll();
		    }
		}
		i++;
	    }
	    if(!found) {
		printf("wn: invalid option : %s\n",argv[j]);
		errcount++;
	    }
	}
    }
#ifdef LABONLY
    /* reset WNDEBUG as it was before wn was run */
    if (wndebug)
	putenv("WNDEBUG=on");
    else
	putenv("WNDEBUG=");
#endif
    exit(errcount);
}

static printsearches(word, dbase,  search)
char *word;
int dbase;
unsigned long search;
{
    int j;

    printf("\nInformation available for %s %s\n", partnames[dbase], word);
    for (j = 1; j <= MAXSEARCH; j++)
	if ((search & bit(j)) && searchstr[j].option) {
	    printf("\t");
	    printf(searchstr[j].template,
		   partchars[dbase], partchars[dbase]);
	    printf(searchstr[j].helpstr);
	    printf("\n");
	}
}

/*
  Revision log:
  
  $Log:	wn.c,v $
 * Revision 1.23  92/04/07  16:39:59  wn
 * moved freeAll() calls
 * 
 * Revision 1.22  92/03/26  15:25:11  wn
 * *** empty log message ***
 * 
 * Revision 1.21  92/03/26  14:48:37  wn
 * changed to call is_defined (removed ISDEF and -isdf{n|v|a} options)
 * 
 * Revision 1.20  92/03/24  14:58:03  wn
 * removed -bracn & -bracv
 * 
 * Revision 1.19  92/03/19  16:36:30  wn
 * fixed typo - sprtn
 * 
 * Revision 1.18  92/03/19  11:48:48  wn
 * change options struct
 * 
  * Revision 1.17  92/03/11  11:26:12  wn
  * added -dbg flag for internal use
  * 
  * Revision 1.16  92/03/06  16:09:10  wn
  * renamed some options, added -hholn, set dflag = 0
  * 
  * Revision 1.15  92/02/21  16:27:52  wn
  * changed search.h to wn.h
  * 
  * Revision 1.14  92/02/05  12:58:26  wn
  * changed to use new morphstr instead of morphit
  * 
  * Revision 1.13  92/01/22  10:41:05  wn
  * init dflag to 0
  * 
  * Revision 1.12  92/01/03  11:51:57  wn
  * added call to strsubst() to replace spaces w/underscore before calling
  * findtheinfo()
  * 
  * Revision 1.11  91/12/18  11:54:32  wn
  * added include for search.h
  * 
  * Revision 1.10  91/12/13  15:44:03  wn
  * changed include file names and pass morphit char as 2nd arg instead
  * of string
  * 
  * Revision 1.9  91/12/04  10:48:24  wn
  * search on original word, then morphed word.
  * also some cleanup
  * 
  * Revision 1.8  91/11/26  17:00:51  wn
  * added "-l" option to print license
  * 
  * Revision 1.7  91/11/22  14:54:25  wn
  * added petainyms to list of valid search types
  * 
  * Revision 1.6  91/11/06  16:47:20  wn
  * added missing help messages and merged help struct into options
  * 
  * Revision 1.5  91/09/25  15:31:20  wn
  * created options structure from info that used to be in enums.c
  * and change code to use structure.
  * 
  * Revision 1.4  91/09/25  09:47:58  wn
  * removed include for morph.h
  * 
  * Revision 1.3  91/09/17  16:35:08  wn
  * changed static int dflag back to int dflag
  * 
  * Revision 1.2  91/09/17  16:33:32  wn
  * changed find.new.h to find.h
  * 
  * Revision 1.1  91/09/17  15:51:09  wn
  * Initial revision
  * 
  */
