/*

  wn.h - header file for use with WordNet search code and morphology 
         function

*/

/* $Id: wn.h,v 1.10 92/03/18 11:51:04 wn Exp $ */

/* Data structures used by search code and returned by
   index_lookup(), read_synset() and findtheinfo_ds() */

/* Structure for index file - returned from index_lookup() */

typedef struct {
    char *wd;			/* word string */
    char *pos;			/* part of speech */
    int sense_cnt;		/* sense (collins) count */
    int off_cnt;		/* number of offsets */
    long *offset;		/* offsets of synsets containing word */
    int ptruse_cnt;		/* number of pointers used */
    int *ptruse;		/* pointers used */
} Index, *IndexPtr;

/* Structure for data file synset - returned from read_synset()
   and findtheinfo_ds() */

typedef struct ss {
    long hereiam;		/* current file position */
    int sstype;			/* type of ADJ synset */
    int fnum;			/* file number that synset comes from */
    char *pos;			/* part of speech */
    int wcount;			/* number of words in synset */
    char **words;		/* words in synset */
    int *wnsense;		/* sense number in wordnet */
    int whichword;		/* which word in synset we're looking for */
    int ptrcount;		/* number of pointers */
    int *ptrtyp;		/* pointer types */
    long *ptroff;		/* pointer offsets */
    int *ppos;			/* pointer part of speech */
    int *pto;			/* pointer 'to' fields */
    int *pfrm;			/* pointer 'from' fields */
    int fcount;			/* number of verb frames */
    int *frmid;			/* frame numbers */
    int *frmto;			/* frame 'to' fields */
    char *defn;			/* synset gloss (definition) */

    /* these fields are used if a data structure is returned
       instead of a text buffer */

    struct ss *nextss;		/* ptr to next synset containing searchword */
    int searchtype;		/* type of search performed */
    struct ss *ptrlist;		/* ptr to synset list result of search */
} Synset, *SynsetPtr;

/* Global variables */

extern int SenseCount;		/* number of senses word has */
extern int OutSenseCount;	/* number of senses search found */
extern int dflag;		/* if set, print definitional glosses */
extern char *(*wn_filenames[4])[]; /* names of lex files ground in database */

/* Search and database functions */

extern char *findtheinfo();	/* search algorithm */
extern SynsetPtr findtheinfo_ds(); /* search algorithm, returns data struct */
extern unsigned long is_defined(); /* look for word in WordNet database */
extern IndexPtr index_lookup(); /* find word in index file */
extern SynsetPtr read_synset();	/* read synset from data file */
extern void freeAll();

/* Morphology functions */

extern char *morphword();	/* morphology function for a word */
extern char *morphstr();	/* morphology function for a string */

/* Utility functions */

extern char *bin_search();	/* binary search routine */
extern int cntwords();		/* number of words in a string */
extern char *strtolower();	/* convert string to lower case */
extern char *strsubst();	/* replace all occurrences of
				   'from' with 'to' in 'str' */
/*
  Revision log:

  $Log:	wn.h,v $
 * Revision 1.10  92/03/18  11:51:04  wn
 * added bin_search
 * 
 * Revision 1.9  92/03/17  11:27:20  wn
 * changed is_defined from int to unsigned long
 * 
 * Revision 1.8  92/03/10  10:59:25  wn
 * added extern for dflag
 * 
 * Revision 1.7  92/03/04  15:00:31  wn
 * made data structures global
 * 
 * Revision 1.6  92/02/25  12:35:47  wn
 * no changes
 * 
 * Revision 1.5  92/02/21  16:28:33  wn
 * renamed from search.h and added cntwords()
 * 
 * Revision 1.4  92/02/12  11:06:52  wn
 * removed prototyping from declarations
 * 
 * Revision 1.3  92/02/12  10:34:59  wn
 * added strtolower() and strsubst()
 * 
 * Revision 1.2  92/02/05  11:05:35  wn
 * added morphstr() and changed morphit() to morphword()
 * 
 * Revision 1.1  92/01/22  10:43:39  wn
 * Initial revision
 * 
*/
